@@ -2,6 +2,7 @@
 """
 import os
 import json
+from typing import List, Optional
 import requests
 import re
 from datetime import datetime
@@ -15,13 +16,14 @@ OBJECT_ID_RE = r".*\/md\/.*\/obj\/(\d+)"
 COOLDOWN_SECONDS = 20
 
 
-class RetrySession():
-    def __init__(self, logger = None):
+class RetrySession:
+    def __init__(self, logger=None):
         self.session = requests.Session()
         if logger is None:
             self.logger = getLogger("RetrySession")
         else:
             self.logger = logger
+
     def get(self, *args, **kwargs):
         retry_count = 0
         retry_timeout = 10
@@ -38,6 +40,7 @@ class RetrySession():
                     retry_timeout = 30
                 self.logger.error("Retry # %d in %d seconds", retry_count, retry_timeout)
                 sleep(retry_timeout)
+
     def post(self, *args, **kwargs):
         retry_count = 0
         retry_timeout = 10
@@ -119,6 +122,39 @@ class ExportJob:
         return False
 
 
+# TODO: Refactor, code repetition
+class ExecutionPoller:
+    def __init__(self, gd, poll_url):
+        self.gd = gd
+        self.poll_url = poll_url
+        self.done = False
+        self.status = None
+
+    def update_status(self):
+        """Update status and return .done attribute value"""
+        url = _url(self.gd.domain, self.poll_url.lstrip("/"))
+        response = self.gd.session.get(url, headers=self.gd.headers)
+        status = response.json().get("execution", {}).get("status")
+        if status is None:
+            raise Exception(response.text)
+        self.status = status
+        if self.status == "DONE":
+            self.done = True
+        elif self.status == "RUNNING":
+            self.done = False
+        else:
+            raise Exception(f"Execution status is prohibited: {self.status}")
+        return self.done
+
+    def wait(self, interval=5):
+        """Wait until the task is done"""
+        if self.done:
+            return True
+        while not self.done:
+            self.update_status()
+            sleep(interval)
+
+
 class GoodData:
     """Main class for interacting with GoodData API."""
 
@@ -315,19 +351,32 @@ class GoodData:
             }
         }
         response = self.session.post(url, json=data, headers=self.headers)
-        while response.status_code != 200:
+        try:
+            target_link = response.json()["uri"][1:]
+        except json.decoder.JSONDecodeError as error:
+            self.logger.error(error)
+            self.logger.error(response.text)
+            self.logger.error("status_code: %d", response.status_code)
+            raise error
+        url = _url(self.domain, target_link)
+
+        attempts_count = 60
+        while attempts_count:
             sleep(1)
-            try:
-                target_link = response.json()["uri"][1:]
-            except json.decoder.JSONDecodeError as error:
-                self.logger.error(error)
-                self.logger.error(response.text)
-                self.logger.error("status_code: %d", response.status_code)
-                raise error
-            url = _url(self.domain, target_link)
             response = self.session.get(url, headers=self.headers)
-        with open(fname, "w", encoding="utf-8") as fid:
-            fid.write(response.text)
+            if response.status_code == 200:
+                self.logger.info("Report returned, status_code: %d", response.status_code)
+                break
+            elif response.status_code == 204:
+                self.logger.info("Empty report returned, status_code: %d", response.status_code)
+                break
+            else:
+                self.logger.info("Report not ready yet, continue polling, status_code: %d", response.status_code)
+            attempts_count -= 1
+
+        if response.text.strip():
+            with open(fname, "w", encoding="utf-8") as fid:
+                fid.write(response.text)
 
     def get_user(self, user_link: str):
         """Get user details. `user_link` may be either whole link (starting /gdc/account/profile) or user id
@@ -650,3 +699,41 @@ class GoodData:
             return found[0]
         else:
             return None
+
+    def list_schedules(self, project_id=None):
+        """Get schedules for project"""
+        project_id = project_id or self.pid
+        self.login()
+        url = _url(self.domain, f"gdc/projects/{project_id}/schedules")
+        response = self.session.get(url, headers=self.headers)
+        return response.json().get("schedules", {}).get("items", [])
+
+    def get_schedule(self, schedule_id, project_id=None):
+        """Get schedule by its id"""
+        project_id = project_id or self.pid
+        self.login()
+        url = _url(self.domain, f"gdc/projects/{project_id}/schedules/{schedule_id}")
+        response = self.session.get(url, headers=self.headers)
+        return response.json().get("schedule", {})
+
+    def execute_schedule(self, schedule_id, datasets: Optional[List] = None, project_id=None):
+        """Execute a schedule and return polling object"""
+        project_id = project_id or self.pid
+        self.login()
+        url = _url(self.domain, f"gdc/projects/{project_id}/schedules/{schedule_id}/executions")
+        sdata = self.get_schedule(schedule_id, project_id)
+        if not datasets:
+            params = sdata.get("params", {})
+        else:
+            dset_lst = [{"dataset": dset, "uploadMode": "DEFAULT"} for dset in datasets]
+            params = {"GDC_DATALOAD_DATASETS": json.dumps(dset_lst), "GDC_DATALOAD_SINGLE_RUN_LOAD_MODE": "DEFAULT"}
+        payload = {"execution": {"params": params}}
+        response = self.session.post(url, headers=self.headers, json=payload)
+        link = response.json().get("execution", {}).get("links", {}).get("self", None)
+        return ExecutionPoller(self, link)
+
+    def get_schedule_execution(self, schedule_id, execution_id, project_id=None):
+        """Get schedule execution status"""
+        project_id = project_id or self.pid
+        self.login()
+        url = _url(self.domain, f"gdc/projects/{project_id}/schedules/{schedule_id}/executions/{execution_id}")
